from . import COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import DirectDateTimeField, StringField


class ReferencedDocument(Element):
    issue_date_time = DirectDateTimeField(NS_RAM, "IssueDateTime", required=False,
                                          profile=COMFORT)
    id = StringField(NS_RAM, "ID", required=False,
                     profile=COMFORT)


class BuyerOrderReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "BuyerOrderReferencedDocument"


class ContractReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "ContractReferencedDocument"


class AdditionalReferencedDocument(Element):
    issue_date_time = DirectDateTimeField(NS_RAM, "IssueDateTime", required=False,
                                          profile=COMFORT)
    type_code = StringField(NS_RAM, "TypeCode", profile=EXTENDED, required=True)
    id = StringField(NS_RAM, "ID", required=False,
                     profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "AdditionalReferencedDocument"


class CustomerOrderReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "CustomerOrderReferencedDocument"


class DespatchAdviceReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "DespatchAdviceReferencedDocument"


class LineCustomerOrderReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "CustomerOrderReferencedDocument"


class LineBuyerOrderReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "BuyerOrderReferencedDocument"


class LineContractReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "ContractReferencedDocument"


class LineDespatchAdviceReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "DespatchAdviceReferencedDocument"


class LineReceivingAdviceReferencedDocument(ReferencedDocument):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "ReceivingAdviceReferencedDocument"


class LineAdditionalReferencedDocument(Element):
    line_id = StringField(NS_RAM, "LineID", required=False, profile=EXTENDED)
    issue_date_time = DirectDateTimeField(NS_RAM, "IssueDateTime", required=False,
                                          profile=COMFORT)
    type_code = StringField(NS_RAM, "TypeCode", profile=EXTENDED, required=True)
    id = StringField(NS_RAM, "ID", required=False,
                     profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "AdditionalReferencedDocument"


class DeliveryNoteReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "DeliveryNoteReferencedDocument"


class LineDeliveryNoteReferencedDocument(ReferencedDocument):
    class Meta:
        namespace = NS_RAM
        tag = "DeliveryNoteReferencedDocument"



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .accounting import (AccountingAccount, ApplicableTradeTax,
                         BillingSpecifiedPeriod, TradeAllowanceCharge)
from .delivery import SupplyChainEvent
from .elements import Element
from .fields import (CurrencyField, Field, MultiField, QuantityField,
                     StringField)
from .note import IncludedNote
from .party import ShipToTradeParty, UltimateShipToTradeParty
from .product import TradeProduct
from .references import (LineAdditionalReferencedDocument,
                         LineBuyerOrderReferencedDocument,
                         LineContractReferencedDocument,
                         LineCustomerOrderReferencedDocument,
                         LineDeliveryNoteReferencedDocument,
                         LineDespatchAdviceReferencedDocument,
                         LineReceivingAdviceReferencedDocument)


class AllowanceCharge(TradeAllowanceCharge):
    class Meta:
        namespace = NS_RAM
        tag = "AppliedTradeAllowanceCharge"


class GrossPrice(Element):
    amount = CurrencyField(NS_RAM, "ChargeAmount", required=True, profile=COMFORT,
                           _d="Bruttopreis")
    basis_quantity = QuantityField(NS_RAM, "BasisQuantity", required=False,
                                   profile=COMFORT, _d="Preisbasismenge")
    charge = MultiField(AllowanceCharge, required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "GrossPriceProductTradePrice"


class NetPrice(Element):
    amount = CurrencyField(NS_RAM, "ChargeAmount", required=True, profile=COMFORT)
    basis_quantity = QuantityField(NS_RAM, "BasisQuantity", required=False,
                                   profile=COMFORT, _d="Preisbasismenge")

    class Meta:
        namespace = NS_RAM
        tag = "NetPriceProductTradePrice"


class LineDocument(Element):
    line_id = StringField(NS_RAM, "LineID")
    notes = MultiField(IncludedNote)

    class Meta:
        namespace = NS_RAM
        tag = "AssociatedDocumentLineDocument"


class LineAgreement(Element):
    buyer_order = Field(LineBuyerOrderReferencedDocument, required=False, profile=EXTENDED)
    contract = Field(LineContractReferencedDocument, required=False, profile=EXTENDED)
    customer_order = Field(LineCustomerOrderReferencedDocument, required=False, profile=EXTENDED)
    additional_references = MultiField(LineAdditionalReferencedDocument, required=False,
                                       profile=COMFORT)
    gross = Field(GrossPrice, required=False, profile=COMFORT)
    net = Field(NetPrice)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedSupplyChainTradeAgreement"


class LineDelivery(Element):
    billed_quantity = QuantityField(NS_RAM, "BilledQuantity", required=True,
                                    profile=BASIC, _d="Menge, berechnet")
    charge_free_quantity = QuantityField(NS_RAM, "ChargeFreeQuantity", required=False,
                                         profile=EXTENDED, _d="Menge, ohne Berechnung")
    package_quantity = QuantityField(NS_RAM, "PackageQuantity", required=False,
                                     profile=EXTENDED, _d="Anzahl Packstücke")
    ship_to = Field(ShipToTradeParty, required=False, profile=EXTENDED)
    ultimate_ship_to = Field(UltimateShipToTradeParty, required=False, profile=EXTENDED)
    event = Field(SupplyChainEvent, required=False, profile=EXTENDED,
                  _d="Detailinformationen zur tatsächlichen Lieferung")
    despatch_advice = Field(LineDespatchAdviceReferencedDocument, required=False,
                            profile=EXTENDED)
    receiving_advice = Field(LineReceivingAdviceReferencedDocument, required=False,
                             profile=EXTENDED)
    delivery_note = Field(LineDeliveryNoteReferencedDocument, required=False,
                          profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedSupplyChainTradeDelivery"


class LineSummation(Element):
    total_amount = CurrencyField(NS_RAM, "LineTotalAmount", required=True,
                                 profile=COMFORT)
    total_allowance_charge = CurrencyField(NS_RAM, "TotalAllowanceChargeAmount",
                                           required=False, profile=EXTENDED, _d="Gesamtbetrag der Zu- und Abschläge")

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeSettlementMonetarySummation"


class LineSettlement(Element):
    trade_tax = Field(ApplicableTradeTax, required=False, profile=COMFORT)
    period = Field(BillingSpecifiedPeriod, required=False, profile=EXTENDED)
    accounting_account = Field(AccountingAccount, required=False, profile=EXTENDED,
                               _d="Kostenstelle")
    monetary_summation = Field(LineSummation, required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedSupplyChainTradeSettlement"


class LineItem(Element):
    document = Field(LineDocument, required=True)
    agreement = Field(LineAgreement)
    delivery = Field(LineDelivery)
    settlement = Field(LineSettlement, required=True)
    product = Field(TradeProduct)

    class Meta:
        namespace = NS_RAM
        tag = "IncludedSupplyChainTradeLineItem"



import xml.etree.cElementTree as ET

from drafthorse.models.note import IncludedNote

from . import BASIC, EXTENDED, NS_RAM, NS_UDT, NS_FERD_1p0
from .elements import Element
from .fields import (DateTimeField, Field, IndicatorField, MultiField,
                     MultiStringField, StringField)
from .trade import TradeTransaction


class GuidelineDocumentContextParameter(Element):
    id = StringField(NS_RAM, "ID")

    class Meta:
        namespace = NS_RAM
        tag = "GuidelineSpecifiedDocumentContextParameter"


class BusinessDocumentContextParameter(Element):
    id = StringField(NS_RAM, "ID")

    class Meta:
        namespace = NS_RAM
        tag = "BusinessProcessSpecifiedDocumentContextParameter"


class DocumentContext(Element):
    test_indicator = IndicatorField(NS_RAM, "TestIndicator", required=False,
                                    profile=BASIC, _d="Testkennzeichen")
    business_parameter = Field(BusinessDocumentContextParameter, required=False,
                               profile=EXTENDED, _d="Geschäftsprozess, Wert")
    guideline_parameter = Field(GuidelineDocumentContextParameter, required=True,
                                profile=BASIC, _d="Anwendungsempfehlung")

    class Meta:
        namespace = NS_FERD_1p0
        tag = "SpecifiedExchangedDocumentContext"


class EffectivePeriod(Element):
    complete = DateTimeField(NS_RAM, "CompleteDateTime")

    class Meta:
        namespace = NS_RAM
        tag = "EffectiveSpecifiedPeriod"


class Header(Element):
    id = StringField(NS_RAM, "ID", required=True, profile=BASIC,
                     _d="Rechnungsnummer")
    name = StringField(NS_RAM, "Name", required=True, profile=BASIC,
                       _d="Dokumentenart (Freitext)")
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=BASIC,
                            _d="Dokumentenart (Code)")
    issue_date_time = DateTimeField(NS_RAM, "IssueDateTime", required=True,
                                    profile=BASIC, _d="Rechnungsdatum")
    copy_indicator = IndicatorField(NS_RAM, "CopyIndicator", required=False,
                                    profile=EXTENDED, _d="Indikator Original/Kopie")
    languages = MultiStringField(NS_RAM, "LanguageID", required=False, profile=EXTENDED)
    effective_period = Field(EffectivePeriod, required=False, profile=EXTENDED,
                             _d="Vertragliches Fälligkeitsdatum der Rechnung")
    notes = MultiField(IncludedNote)

    class Meta:
        namespace = NS_FERD_1p0
        tag = "HeaderExchangedDocument"


class Document(Element):
    context = Field(DocumentContext, required=True)
    header = Field(Header, required=True)
    trade = Field(TradeTransaction, required=True)

    def __init__(self):
        super().__init__()
        ET.register_namespace("xsi", "http://www.w3.org/2001/XMLSchema-instance")
        ET.register_namespace("rsm", NS_FERD_1p0)
        ET.register_namespace("ram", NS_RAM)
        ET.register_namespace("udt", NS_UDT)

    class Meta:
        namespace = NS_FERD_1p0
        tag = "CrossIndustryDocument"



class Container:
    def __init__(self, child_type):
        super().__init__()
        self.children = []
        self.child_type = child_type

    def add(self, item):
        if isinstance(self.child_type, type) and not isinstance(item, self.child_type):
            raise TypeError("{} is not of type {}".format(item, self.child_type))
        self.children.append(item)

    def append_to(self, node):
        for child in self.children:
            child.append_to(node)

    def get_tag(self):
        return '{%s}%s' % (self.child_type.Meta.namespace, self.child_type.Meta.tag)

    def empty_element(self):
        return self.child_type()

    def add_from_etree(self, root):
        childel = self.empty_element()
        childel.from_etree(root)
        self.add(childel)


class SimpleContainer(Container):
    def __init__(self, child_type, namespace, tag):
        super().__init__(child_type)
        self.namespace = namespace
        self.tag = tag

    def get_tag(self):
        return '{%s}%s' % (self.namespace, self.tag)

    def empty_element(self):
        raise NotImplementedError()

    def set_element(self, el, child):
        raise NotImplementedError()

    def append_to(self, node):
        for child in self.children:
            el = self.empty_element()
            self.set_element(el, child)
            el.append_to(node)

    def add_from_etree(self, root):
        self.add(root.text)


class CurrencyContainer(SimpleContainer):

    def empty_element(self):
        from .elements import CurrencyElement
        return CurrencyElement(namespace=self.namespace, tag=self.tag)

    def set_element(self, el, child):
        el.amount = child[0]
        el.currency = child[1]

    def add_from_etree(self, root):
        self.add((root.text, root.attrib['currencyID']))


class IDContainer(SimpleContainer):
    def empty_element(self):
        from .elements import IDElement
        return IDElement(namespace=self.namespace, tag=self.tag)

    def set_element(self, el, child):
        el.text = child[1]
        el.scheme_id = child[0]

    def add_from_etree(self, root):
        self.add((root.attrib['schemeID'], root.text))


class StringContainer(SimpleContainer):

    def empty_element(self):
        from .elements import StringElement
        return StringElement(namespace=self.namespace, tag=self.tag)

    def set_element(self, el, child):
        el.text = child

    def add_from_etree(self, root):
        self.add(root.text)



NS_FERD_1p0 = "urn:ferd:CrossIndustryDocument:invoice:1p0"
NS_UDT = "urn:un:unece:uncefact:data:standard:UnqualifiedDataType:15"
NS_RAM = "urn:un:unece:uncefact:data:standard:ReusableAggregateBusinessInformationEntity:12"
BASIC = "BASIC"
COMFORT = "COMFORT"
EXTENDED = "EXTENDED"



# Based on code of the factur-x Python package:
# Copyright 2016-2018, Alexis de Lattre <alexis.delattre@akretion.com>
# Modified for ZUGFeRD and Python 3 by Raphael Michel, 2018
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * The name of the authors may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import datetime
import hashlib
from PyPDF2 import PdfFileReader, PdfFileWriter
from PyPDF2.generic import createStringObject, DictionaryObject, NameObject, DecodedStreamObject, ArrayObject
from io import BytesIO
from lxml import etree


def attach_xml(original_pdf, xml_data, level='BASIC'):
    if not isinstance(original_pdf, bytes):
        raise TypeError("Please supply original PDF as bytes.")
    if not isinstance(xml_data, bytes):
        raise TypeError("Please supply XML data as bytes.")

    reader = PdfFileReader(BytesIO(original_pdf))
    output = PdfFileWriter()
    # for page in reader.pages:
    #    output.addPage(page)

    output._header = "%PDF-1.6".encode()
    output.appendPagesFromReader(reader)

    original_pdf_id = reader.trailer.get('/ID')
    if original_pdf_id:
        output._ID = original_pdf_id
        # else : generate some ?

    _facturx_update_metadata_add_attachment(
        output, xml_data, {}, level,
        output_intents=_get_original_output_intents(reader),
    )

    outbuffer = BytesIO()
    output.write(outbuffer)
    outbuffer.seek(0)
    return outbuffer


def _get_original_output_intents(original_pdf):
    output_intents = []
    try:
        pdf_root = original_pdf.trailer['/Root']
        ori_output_intents = pdf_root['/OutputIntents']
        for ori_output_intent in ori_output_intents:
            ori_output_intent_dict = ori_output_intent.getObject()
            dest_output_profile_dict = \
                ori_output_intent_dict['/DestOutputProfile'].getObject()
            output_intents.append(
                (ori_output_intent_dict, dest_output_profile_dict))
    except:
        pass
    return output_intents


def _prepare_pdf_metadata_txt(pdf_metadata):
    pdf_date = datetime.datetime.now().isoformat()
    info_dict = {
        '/Author': pdf_metadata.get('author', ''),
        '/CreationDate': pdf_date,
        '/Creator': 'pretix',
        '/Keywords': pdf_metadata.get('keywords', ''),
        '/ModDate': pdf_date,
        '/Subject': pdf_metadata.get('subject', ''),
        '/Title': pdf_metadata.get('title', ''),
    }
    return info_dict


def _prepare_pdf_metadata_xml(level, pdf_metadata):
    nsmap_x = {'x': 'adobe:ns:meta/'}
    nsmap_rdf = {'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'}
    nsmap_dc = {'dc': 'http://purl.org/dc/elements/1.1/'}
    nsmap_pdf = {'pdf': 'http://ns.adobe.com/pdf/1.3/'}
    nsmap_xmp = {'xmp': 'http://ns.adobe.com/xap/1.0/'}
    nsmap_pdfaid = {'pdfaid': 'http://www.aiim.org/pdfa/ns/id/'}
    nsmap_zf = {'zf': 'urn:ferd:pdfa:CrossIndustryDocument:invoice:1p0#'}
    ns_x = '{%s}' % nsmap_x['x']
    ns_dc = '{%s}' % nsmap_dc['dc']
    ns_rdf = '{%s}' % nsmap_rdf['rdf']
    ns_pdf = '{%s}' % nsmap_pdf['pdf']
    ns_xmp = '{%s}' % nsmap_xmp['xmp']
    ns_pdfaid = '{%s}' % nsmap_pdfaid['pdfaid']
    ns_zf = '{%s}' % nsmap_zf['zf']
    ns_xml = '{http://www.w3.org/XML/1998/namespace}'

    root = etree.Element(ns_x + 'xmpmeta', nsmap=nsmap_x)
    rdf = etree.SubElement(root, ns_rdf + 'RDF', nsmap=nsmap_rdf)
    desc_pdfaid = etree.SubElement(rdf, ns_rdf + 'Description', nsmap=nsmap_pdfaid)
    desc_pdfaid.set(ns_rdf + 'about', '')
    etree.SubElement(desc_pdfaid, ns_pdfaid + 'part').text = '3'
    etree.SubElement(desc_pdfaid, ns_pdfaid + 'conformance').text = 'B'
    desc_dc = etree.SubElement(rdf, ns_rdf + 'Description', nsmap=nsmap_dc)
    desc_dc.set(ns_rdf + 'about', '')
    dc_title = etree.SubElement(desc_dc, ns_dc + 'title')
    dc_title_alt = etree.SubElement(dc_title, ns_rdf + 'Alt')
    dc_title_alt_li = etree.SubElement(dc_title_alt, ns_rdf + 'li')
    dc_title_alt_li.text = pdf_metadata.get('title', '')
    dc_title_alt_li.set(ns_xml + 'lang', 'x-default')
    dc_creator = etree.SubElement(desc_dc, ns_dc + 'creator')
    dc_creator_seq = etree.SubElement(dc_creator, ns_rdf + 'Seq')
    etree.SubElement(dc_creator_seq, ns_rdf + 'li').text = pdf_metadata.get('author', '')
    dc_desc = etree.SubElement(desc_dc, ns_dc + 'description')
    dc_desc_alt = etree.SubElement(dc_desc, ns_rdf + 'Alt')
    dc_desc_alt_li = etree.SubElement(dc_desc_alt, ns_rdf + 'li')
    dc_desc_alt_li.text = pdf_metadata.get('subject', '')
    dc_desc_alt_li.set(ns_xml + 'lang', 'x-default')
    desc_adobe = etree.SubElement(rdf, ns_rdf + 'Description', nsmap=nsmap_pdf)
    desc_adobe.set(ns_rdf + 'about', '')
    producer = etree.SubElement(desc_adobe, ns_pdf + 'Producer')
    producer.text = 'PyPDF2'
    desc_xmp = etree.SubElement(rdf, ns_rdf + 'Description', nsmap=nsmap_xmp)
    desc_xmp.set(ns_rdf + 'about', '')
    creator = etree.SubElement(desc_xmp, ns_xmp + 'CreatorTool')
    creator.text = 'pretix'
    timestamp = datetime.datetime.now().isoformat()
    etree.SubElement(desc_xmp, ns_xmp + 'CreateDate').text = timestamp
    etree.SubElement(desc_xmp, ns_xmp + 'ModifyDate').text = timestamp

    """
    xmp_file = resource_filename(__name__, 'xmp/Factur-X_extension_schema.xmp')
    # Reason for defining a parser below:
    # http://lxml.de/FAQ.html#why-doesn-t-the-pretty-print-option-reformat-my-xml-output
    parser = etree.XMLParser(remove_blank_text=True)
    facturx_ext_schema_root = etree.parse(open(xmp_file), parser)
    # The Factur-X extension schema must be embedded into each PDF document
    facturx_ext_schema_desc_xpath = facturx_ext_schema_root.xpath(
        '//rdf:Description', namespaces=nsmap_rdf)
    rdf.append(facturx_ext_schema_desc_xpath[1])
    """
    # Now is the ZUGFeRD description tag
    zugferd_desc = etree.SubElement(rdf, ns_rdf + 'Description', nsmap=nsmap_zf)
    zugferd_desc.set(ns_rdf + 'about', '')
    fx_doc_type = etree.SubElement(zugferd_desc, ns_zf + 'DocumentType', nsmap=nsmap_zf)
    fx_doc_type.text = 'INVOICE'
    fx_doc_filename = etree.SubElement(zugferd_desc, ns_zf + 'DocumentFileName', nsmap=nsmap_zf)
    fx_doc_filename.text = 'ZUGFeRD-invoice.xml'
    fx_doc_version = etree.SubElement(zugferd_desc, ns_zf + 'Version', nsmap=nsmap_zf)
    fx_doc_version.text = '1.0'
    fx_conformance_level = etree.SubElement(zugferd_desc, ns_zf + 'ConformanceLevel', nsmap=nsmap_zf)
    fx_conformance_level.text = level

    # TODO: should be UTF-16be ??
    xml_str = etree.tostring(root, pretty_print=True, encoding="UTF-8", xml_declaration=False)
    head = u'<?xpacket begin="\ufeff" id="W5M0MpCehiHzreSzNTczkc9d"?>'.encode('utf-8')
    tail = u'<?xpacket end="w"?>'.encode('utf-8')
    xml_final_str = head + xml_str + tail
    return xml_final_str


def _facturx_update_metadata_add_attachment(pdf_filestream, facturx_xml_str, pdf_metadata, facturx_level,
                                            output_intents):
    md5sum = hashlib.md5(facturx_xml_str).hexdigest()
    md5sum_obj = createStringObject(md5sum)
    params_dict = DictionaryObject({
        NameObject('/CheckSum'): md5sum_obj,
        NameObject('/ModDate'): createStringObject(datetime.datetime.now().isoformat()),
        NameObject('/Size'): NameObject(str(len(facturx_xml_str))),
    })
    file_entry = DecodedStreamObject()
    file_entry.setData(facturx_xml_str)  # here we integrate the file itself
    file_entry.update({
        NameObject("/Type"): NameObject("/EmbeddedFile"),
        NameObject("/Params"): params_dict,
        # 2F is '/' in hexadecimal
        NameObject("/Subtype"): NameObject("/text#2Fxml"),
    })
    file_entry_obj = pdf_filestream._addObject(file_entry)
    # The Filespec entry
    ef_dict = DictionaryObject({
        NameObject("/F"): file_entry_obj,
        NameObject('/UF'): file_entry_obj,
    })

    fname_obj = createStringObject("ZUGFeRD-invoice.xml")
    filespec_dict = DictionaryObject({
        NameObject("/AFRelationship"): NameObject("/Data"),
        NameObject("/Desc"): createStringObject("Factur-X Invoice"),
        NameObject("/Type"): NameObject("/Filespec"),
        NameObject("/F"): fname_obj,
        NameObject("/EF"): ef_dict,
        NameObject("/UF"): fname_obj,
    })
    filespec_obj = pdf_filestream._addObject(filespec_dict)
    name_arrayobj_cdict = {fname_obj: filespec_obj}
    name_arrayobj_content_sort = list(
        sorted(name_arrayobj_cdict.items(), key=lambda x: x[0]))
    name_arrayobj_content_final = []
    af_list = []
    for (fname_obj, filespec_obj) in name_arrayobj_content_sort:
        name_arrayobj_content_final += [fname_obj, filespec_obj]
        af_list.append(filespec_obj)
    embedded_files_names_dict = DictionaryObject({
        NameObject("/Names"): ArrayObject(name_arrayobj_content_final),
    })
    # Then create the entry for the root, as it needs a
    # reference to the Filespec
    embedded_files_dict = DictionaryObject({
        NameObject("/EmbeddedFiles"): embedded_files_names_dict,
    })
    res_output_intents = []
    for output_intent_dict, dest_output_profile_dict in output_intents:
        dest_output_profile_obj = pdf_filestream._addObject(
            dest_output_profile_dict)
        # TODO detect if there are no other objects in output_intent_dest_obj
        # than /DestOutputProfile
        output_intent_dict.update({
            NameObject("/DestOutputProfile"): dest_output_profile_obj,
        })
        output_intent_obj = pdf_filestream._addObject(output_intent_dict)
        res_output_intents.append(output_intent_obj)
    # Update the root
    metadata_xml_str = _prepare_pdf_metadata_xml(facturx_level, pdf_metadata)
    metadata_file_entry = DecodedStreamObject()
    metadata_file_entry.setData(metadata_xml_str)
    metadata_file_entry.update({
        NameObject('/Subtype'): NameObject('/XML'),
        NameObject('/Type'): NameObject('/Metadata'),
    })
    metadata_obj = pdf_filestream._addObject(metadata_file_entry)
    af_value_obj = pdf_filestream._addObject(ArrayObject(af_list))
    pdf_filestream._root_object.update({
        NameObject("/AF"): af_value_obj,
        NameObject("/Metadata"): metadata_obj,
        NameObject("/Names"): embedded_files_dict,
        # show attachments when opening PDF
        NameObject("/PageMode"): NameObject("/UseAttachments"),
    })
    if res_output_intents:
        pdf_filestream._root_object.update({
            NameObject("/OutputIntents"): ArrayObject(res_output_intents),
        })
    metadata_txt_dict = _prepare_pdf_metadata_txt(pdf_metadata)
    pdf_filestream.addMetadata(metadata_txt_dict)



import collections
import sys
import xml.etree.cElementTree as ET
from collections import OrderedDict
from datetime import datetime
from decimal import Decimal

from drafthorse.utils import validate_xml

from . import NS_UDT
from .container import Container
from .fields import Field


class BaseElementMeta(type):
    @classmethod
    def __prepare__(self, name, bases):
        return collections.OrderedDict()

    def __new__(mcls, name, bases, attrs):
        cls = super(BaseElementMeta, mcls).__new__(mcls, name, bases, attrs)
        fields = list(cls._fields) if hasattr(cls, '_fields') else []
        for attr, obj in attrs.items():
            if isinstance(obj, Field):
                if sys.version_info < (3, 6):
                    obj.__set_name__(cls, attr)
                fields.append(obj)
        cls._fields = fields
        return cls


class Element(metaclass=BaseElementMeta):
    def __init__(self, **kwargs):
        self.required = kwargs.get('required', False)
        self._data = OrderedDict([
            (f.name, f.initialize() if f.default or f.required else None) for f in self._fields
        ])
        for k, v in kwargs.items():
            setattr(self, k, v)

    def _etree_node(self):
        node = ET.Element(self.get_tag())
        if hasattr(self, 'Meta') and hasattr(self.Meta, 'attributes'):
            for k, v in self.Meta.attributes.items():
                node.set(k, v)
        return node

    def to_etree(self):
        node = self._etree_node()
        for k, v in self._data.items():
            if v is not None:
                v.append_to(node)
        return node

    def get_tag(self):
        return "{%s}%s" % (self.Meta.namespace, self.Meta.tag)

    def append_to(self, node):
        el = self.to_etree()
        if self.required or list(el) or el.text:
            node.append(el)

    def serialize(self):
        xml = b"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + ET.tostring(self.to_etree(), "utf-8")
        return validate_xml(xmlout=xml, schema="ZUGFeRD1p0")

    def from_etree(self, root):
        if hasattr(self, 'Meta') and hasattr(self.Meta, 'namespace') and root.tag != "{%s}%s" % (
                self.Meta.namespace, self.Meta.tag):
            raise TypeError("Invalid XML, found tag {} where {} was expected".format(root.tag, "{%s}%s" % (
                self.Meta.namespace, self.Meta.tag)))
        field_index = {}
        for field in self._fields:
            element = getattr(self, field.name)
            field_index[element.get_tag()] = (field.name, element)
        for child in root:
            if child.tag == ET.Comment:
                continue
            if child.tag in field_index:
                name, childel = field_index[child.tag]
                if isinstance(getattr(self, name), Container):
                    getattr(self, name).add_from_etree(child)
                else:
                    getattr(self, name).from_etree(child)
            else:
                raise TypeError("Unknown element {}".format(child.tag))
        return self

    @classmethod
    def parse(cls, xmlinput):
        from lxml import etree
        root = etree.fromstring(xmlinput)
        return cls().from_etree(root)


class StringElement(Element):
    def __init__(self, namespace, tag, text=""):
        super().__init__()
        self.namespace = namespace
        self.tag = tag
        self.text = text

    def __repr__(self):
        return '<{}: {}>'.format(type(self).__name__, str(self))

    def __str__(self):
        return self.text

    def get_tag(self):
        return "{%s}%s" % (self.namespace, self.tag)

    def to_etree(self):
        node = self._etree_node()
        node.text = self.text
        return node

    def from_etree(self, root):
        self.text = root.text
        return self


class DecimalElement(StringElement):
    def __init__(self, namespace, tag, value=None):
        super().__init__(namespace, tag)
        self.value = value

    def to_etree(self):
        node = self._etree_node()
        node.text = str(self.value) if self.value is not None else ""
        return node

    def __str__(self):
        return self.value

    def from_etree(self, root):
        self.value = Decimal(root.text)
        return self


class QuantityElement(StringElement):
    def __init__(self, namespace, tag, amount="", unit_code=""):
        super().__init__(namespace, tag)
        self.amount = amount
        self.unit_code = unit_code

    def to_etree(self):
        node = self._etree_node()
        node.text = str(self.amount)
        node.attrib["unitCode"] = self.unit_code
        return node

    def __str__(self):
        return "{} {}".format(self.amount, self.unit_code)

    def from_etree(self, root):
        self.amount = Decimal(root.text)
        self.unit_code = root.attrib['unitCode']
        return self


class CurrencyElement(StringElement):
    def __init__(self, namespace, tag, amount="", currency="EUR"):
        super().__init__(namespace, tag)
        self.amount = amount
        self.currency = currency

    def to_etree(self):
        node = self._etree_node()
        node.text = str(self.amount)
        node.attrib["currencyID"] = self.currency
        return node

    def from_etree(self, root):
        self.amount = Decimal(root.text)
        self.currency = root.attrib['currencyID']
        return self

    def __str__(self):
        return "{} {}".format(self.amount, self.currency)


class ClassificationElement(StringElement):
    def __init__(self, namespace, tag, text="", list_id="", list_version_id=""):
        super().__init__(namespace, tag)
        self.text = text
        self.list_id = list_id
        self.list_version_id = list_version_id

    def to_etree(self):
        node = self._etree_node()
        node.text = self.text
        node.attrib['listID'] = self.list_id
        node.attrib['listVersionID'] = self.list_version_id
        return node

    def from_etree(self, root):
        self.text = Decimal(root.text)
        self.list_id = root.attrib['listID']
        self.list_version_id = root.attrib['listVersionID']
        return self

    def __str__(self):
        return "{} ({} {})".format(self.text, self.list_id, self.list_version_id)


class AgencyIDElement(StringElement):
    def __init__(self, namespace, tag, text="", scheme_id=""):
        super().__init__(namespace, tag)
        self.text = text
        self.scheme_id = scheme_id

    def to_etree(self):
        node = self._etree_node()
        node.text = self.text
        node.attrib['schemeAgencyID'] = self.scheme_id
        return node

    def from_etree(self, root):
        self.text = root.text
        self.scheme_id = root.attrib['schemeAgencyID']
        return self

    def __str__(self):
        return "{} ({})".format(self.text, self.scheme_id)


class IDElement(StringElement):
    def __init__(self, namespace, tag, text="", scheme_id=""):
        super().__init__(namespace, tag)
        self.text = text
        self.scheme_id = scheme_id

    def to_etree(self):
        node = self._etree_node()
        node.text = self.text
        node.attrib['schemeID'] = self.scheme_id
        return node

    def from_etree(self, root):
        self.text = root.text
        self.scheme_id = root.attrib['schemeID']
        return self

    def __str__(self):
        return "{} ({})".format(self.text, self.scheme_id)


class DateTimeElement(StringElement):
    def __init__(self, namespace, tag, value=None, format='102'):
        super().__init__(namespace, tag)
        self.value = value
        self.format = format

    def to_etree(self):
        t = self._etree_node()
        node = ET.Element("{%s}%s" % (NS_UDT, "DateTimeString"))
        if self.value:
            if self.format == '102':
                node.text = self.value.strftime("%Y%m%d")
            elif self.format == '616':
                if sys.version_info < (3, 6):
                    node.text = '{}{}'.format(self.value.isocalendar()[0], self.value.isocalendar()[1])
                else:
                    node.text = self.value.strftime("%G%V")
            node.attrib['format'] = self.format
            t.append(node)
        return t

    def from_etree(self, root):
        if len(root) != 1:
            raise TypeError("Date containers should have one child")
        if root[0].tag != "{%s}%s" % (NS_UDT, "DateTimeString"):
            raise TypeError("Tag %s not recognized" % root[0].tag)
        self.format = root[0].attrib['format']
        if self.format == '102':
            self.value = datetime.strptime(root[0].text, '%Y%m%d').date()
        elif self.format == '616':
            if sys.version_info < (3, 6):
                from isoweek import Week
                w = Week(int(root[0].text[:4]), int(root[0].text[4:]))
                self.value = w.monday()
            else:
                self.value = datetime.strptime(root[0].text + '1', '%G%V%u').date()
        else:
            raise TypeError("Date format %s cannot be parsed" % root[0].attrib['format'])
        return self

    def __str__(self):
        return "{}".format(self.value)


class DirectDateTimeElement(StringElement):
    def __init__(self, namespace, tag, value=None):
        super().__init__(namespace, tag)
        self.value = value
        self.format = format

    def to_etree(self):
        t = self._etree_node()
        if self.value:
            t.text = self.value.strftime("%Y-%m-%dT%H:%M:%S")
        return t

    def from_etree(self, root):
        self.value = datetime.strptime(root.text, '%Y-%m-%dT%H:%M:%S').date()
        return self

    def __str__(self):
        return "{}".format(self.value)


class IndicatorElement(StringElement):
    def __init__(self, namespace, tag, value=None):
        super().__init__(namespace, tag)
        self.value = value

    def get_tag(self):
        return "{%s}%s" % (self.namespace, self.tag)

    def to_etree(self):
        t = self._etree_node()
        if self.value is None:
            return t
        node = ET.Element("{%s}%s" % (NS_UDT, "Indicator"))
        node.text = str(self.value).lower()
        t.append(node)
        return t

    def __str__(self):
        return "{}".format(self.value)

    def from_etree(self, root):
        if len(root) != 1:
            raise TypeError("Indicator containers should have one child")
        if root[0].tag != "{%s}%s" % (NS_UDT, "Indicator"):
            raise TypeError("Tag %s not recognized" % root[0].tag)
        self.value = root[0].text == 'true'
        return self



from . import BASIC, EXTENDED, NS_RAM
from .elements import Element
from .fields import DateTimeField, Field, IDField, StringField
from .party import (ShipFromTradeParty, ShipToTradeParty,
                    UltimateShipToTradeParty)
from .references import (DeliveryNoteReferencedDocument,
                         DespatchAdviceReferencedDocument)


class SupplyChainEvent(Element):
    occurrence = DateTimeField(NS_RAM, "OccurrenceDateTime",
                               required=False, profile=BASIC,
                               _d="Tatsächlicher Lieferungszeitpunkt")

    class Meta:
        namespace = NS_RAM
        tag = "ActualDeliverySupplyChainEvent"


class LogisticsTransportMovement(Element):
    mode_code = StringField(NS_RAM, "ModeCode", required=False, profile=EXTENDED)
    id = IDField(NS_RAM, "ID", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLogisticsTransportMovement"


class SupplyChainConsignment(Element):
    movement = Field(LogisticsTransportMovement, required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "RelatedSupplyChainConsignment"


class TradeDelivery(Element):
    consignment = Field(SupplyChainConsignment, default=False, required=False,
                        _d="Detailinformationen zur Konsignation oder Sendung")
    ship_to = Field(ShipToTradeParty, required=False, profile=EXTENDED)
    ultimate_ship_to = Field(UltimateShipToTradeParty, required=False, profile=EXTENDED)
    ship_from = Field(ShipFromTradeParty, required=False, profile=EXTENDED)
    event = Field(SupplyChainEvent, required=False, profile=BASIC)
    despatch_advice = Field(DespatchAdviceReferencedDocument, required=False,
                            profile=EXTENDED)
    delivery_note = Field(DeliveryNoteReferencedDocument, required=False,
                          profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableSupplyChainTradeDelivery"



from . import BASIC, COMFORT, EXTENDED, NS_RAM, NS_FERD_1p0
from .accounting import (ApplicableTradeTax, AppliedTradeTax,
                         BillingSpecifiedPeriod, MonetarySummation,
                         ReceivableAccountingAccount, TradeAllowanceCharge)
from .delivery import TradeDelivery
from .elements import Element
from .fields import CurrencyField, Field, MultiField, StringField
from .party import (BuyerTradeParty, EndUserTradeParty, InvoiceeTradeParty,
                    PayeeTradeParty, SellerTradeParty)
from .payment import PaymentMeans, PaymentTerms
from .references import (AdditionalReferencedDocument,
                         BuyerOrderReferencedDocument,
                         ContractReferencedDocument,
                         CustomerOrderReferencedDocument)
from .tradelines import LineItem


class DeliveryTerms(Element):
    type_code = StringField(NS_RAM, "DeliveryTypeCode", required=False,
                            profile=EXTENDED, _d="Lieferbedingung (Code)")

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradeDeliveryTerms"


class TradeAgreement(Element):
    buyer_reference = StringField(NS_RAM, "BuyerReference", required=False,
                                  profile=COMFORT, _d="Referenz des Käufers")
    seller = Field(SellerTradeParty, required=True, _d="Detailinformationen zum Verkäufer")
    buyer = Field(BuyerTradeParty, required=True)
    end_user = Field(EndUserTradeParty, required=False, _d="Abweichender Endverbraucher")
    delivery_terms = Field(DeliveryTerms, required=False, profile=EXTENDED)
    buyer_order = Field(BuyerOrderReferencedDocument, required=False, profile=COMFORT)
    customer_order = Field(CustomerOrderReferencedDocument, required=False, profile=COMFORT)
    contract = Field(ContractReferencedDocument, required=False, profile=COMFORT)
    additional_references = MultiField(AdditionalReferencedDocument, required=False,
                                       profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableSupplyChainTradeAgreement"


class LogisticsServiceCharge(Element):
    description = StringField(NS_RAM, "Description", required=True, profile=COMFORT,
                              _d="Identifikation der Servicegebühr")
    applied_amount = CurrencyField(NS_RAM, "AppliedAmount", required=True,
                                   profile=COMFORT, _d="Betrag der Servicegebühr")
    trade_tax = MultiField(AppliedTradeTax, required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedLogisticsServiceCharge"


class TradeSettlement(Element):
    payment_reference = StringField(NS_RAM, "PaymentReference")
    currency_code = StringField(NS_RAM, "InvoiceCurrencyCode")
    invoicee = Field(InvoiceeTradeParty, required=False, profile=COMFORT,
                     _d="Rechnungsempfänger")
    payee = Field(PayeeTradeParty, required=False, profile=COMFORT,
                  _d="Zahlungsempfänger")
    payment_means = Field(PaymentMeans)
    trade_tax = MultiField(ApplicableTradeTax)
    period = Field(BillingSpecifiedPeriod, required=False, profile=COMFORT)
    allowance_charge = MultiField(TradeAllowanceCharge, required=False, profile=COMFORT,
                                  _d="Schalter für Zu-/Abschlag")
    service_charge = MultiField(LogisticsServiceCharge, required=False, profile=COMFORT)
    terms = MultiField(PaymentTerms, required=False, profile=COMFORT)
    monetary_summation = Field(MonetarySummation, required=True, profile=BASIC,
                               _d="Detailinformation zu Belegsummen")
    accounting_account = Field(ReceivableAccountingAccount, required=False, profile=EXTENDED,
                               _d="Detailinformationen zur Buchungsreferenz")

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableSupplyChainTradeSettlement"


class TradeTransaction(Element):
    agreement = Field(TradeAgreement, required=True)
    delivery = Field(TradeDelivery, required=True)
    settlement = Field(TradeSettlement, required=True)
    items = MultiField(LineItem, required=True)

    class Meta:
        namespace = NS_FERD_1p0
        tag = "SpecifiedSupplyChainTradeTransaction"



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import (CurrencyField, DateTimeField, DecimalField,
                     IndicatorField, MultiField, QuantityField, StringField)


class LineApplicableTradeTax(Element):
    calculated_amount = CurrencyField(NS_RAM, "CalculatedAmount", required=True,
                                      profile=BASIC, _d="Steuerbetrag")
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=BASIC,
                            _d="Steuerart (Code)")
    exemption_reason = StringField(NS_RAM, "ExemptionReason", required=False,
                                   profile=COMFORT, _d="Grund der Steuerbefreiung (Freitext)")
    category_code = StringField(NS_RAM, "CategoryCode", required=False,
                                profile=COMFORT, _d="Steuerkategorie (Wert)")
    applicable_percent = DecimalField(NS_RAM, "ApplicablePercent",
                                      required=True, profile=BASIC)

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradeTax"


class ApplicableTradeTax(Element):
    calculated_amount = CurrencyField(NS_RAM, "CalculatedAmount", required=True,
                                      profile=BASIC, _d="Steuerbetrag")
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=BASIC,
                            _d="Steuerart (Code)")
    exemption_reason = StringField(NS_RAM, "ExemptionReason", required=False,
                                   profile=COMFORT, _d="Grund der Steuerbefreiung (Freitext)")
    basis_amount = CurrencyField(NS_RAM, "BasisAmount", required=True,
                                 profile=BASIC, _d="Basisbetrag der Steuerberechnung")
    line_total_basis_amount = CurrencyField(NS_RAM, "LineTotalBasisAmount",
                                            required=False, profile=EXTENDED,
                                            _d="Warenbetrag des Steuersatzes")
    allowance_charge_basis_amount = CurrencyField(NS_RAM, "AllowanceChargeBasisAmount",
                                                  required=False, profile=EXTENDED,
                                                  _d="Gesamtbetrag Zu- und Abschläge des Steuersatzes")
    category_code = StringField(NS_RAM, "CategoryCode", required=False,
                                profile=COMFORT, _d="Steuerkategorie (Wert)")
    applicable_percent = DecimalField(NS_RAM, "ApplicablePercent",
                                      required=True, profile=BASIC)

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradeTax"


class AccountingAccount(Element):
    id = StringField(NS_RAM, "ID", required=True, profile=EXTENDED, _d="Buchungsreferenz")

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeAccountingAccount"


class ReceivableAccountingAccount(Element):
    id = StringField(NS_RAM, "ID", required=True, profile=EXTENDED, _d="Buchungsreferenz")

    class Meta:
        namespace = NS_RAM
        tag = "ReceivableSpecifiedTradeAccountingAccount"


class MonetarySummation(Element):
    line_total = CurrencyField(NS_RAM, "LineTotalAmount", required=True,
                               profile=BASIC, _d="Gesamtbetrag der Positionen")
    charge_total = CurrencyField(NS_RAM, "ChargeTotalAmount", required=True,
                                 profile=BASIC, _d="Gesamtbetrag der Zuschläge")
    allowance_total = CurrencyField(NS_RAM, "AllowanceTotalAmount", required=True,
                                    profile=BASIC, _d="Gesamtbetrag der Abschläge")
    tax_basis_total = CurrencyField(NS_RAM, "TaxBasisTotalAmount", required=True,
                                    profile=BASIC, _d="Steuerbasisbetrag")
    tax_total = CurrencyField(NS_RAM, "TaxTotalAmount", required=True,
                              profile=BASIC, _d="Steuergesamtbetrag")
    grand_total = CurrencyField(NS_RAM, "GrandTotalAmount", required=True,
                                profile=BASIC, _d="Bruttosumme")
    prepaid_total = CurrencyField(NS_RAM, "TotalPrepaidAmount", required=False,
                                  profile=COMFORT, _d="Anzahlungsbetrag")
    due_amount = CurrencyField(NS_RAM, "DuePayableAmount", required=False,
                               profile=COMFORT, _d="Zahlbetrag")

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeSettlementMonetarySummation"


class BillingSpecifiedPeriod(Element):
    start = DateTimeField(NS_RAM, "StartDateTime", required=True, profile=COMFORT)
    end = DateTimeField(NS_RAM, "EndDateTime", required=True, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "BillingSpecifiedPeriod"


class AppliedTradeTax(Element):
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=COMFORT)
    category_code = StringField(NS_RAM, "CategoryCode", required=True, profile=COMFORT)
    applicable_percent = StringField(NS_RAM, "ApplicablePercent", required=True, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "AppliedTradeTax"


class CategoryTradeTax(AppliedTradeTax):
    class Meta:
        namespace = NS_RAM
        tag = "CategoryTradeTax"


class TradeAllowanceCharge(Element):
    indicator = IndicatorField(NS_RAM, "ChargeIndicator", required=False, profile=COMFORT,
                               _d="Schalter für Zu-/Abschlag")
    sequence_numeric = DecimalField(NS_RAM, "SequenceNumeric", required=False, profile=EXTENDED,
                                    _d="Berechnungsreihenfolge")
    calculation_percent = DecimalField(NS_RAM, "CalculationPercent",
                                       required=False, profile=EXTENDED,
                                       _d="Rabatt in Prozent")
    basis_amount = CurrencyField(NS_RAM, "BasisAmount", required=False,
                                 profile=EXTENDED, _d="Basisbetrag des Rabatts")
    basis_quantity = QuantityField(NS_RAM, "BasisQuantity", required=False,
                                   profile=EXTENDED, _d="Basismenge des Rabatts")
    actual_amount = CurrencyField(NS_RAM, "ActualAmount", required=True,
                                  profile=COMFORT, _d="Betrag des Zu-/Abschlags")
    reason_code = StringField(NS_RAM, "ReasonCode", required=False, profile=EXTENDED)
    reason = StringField(NS_RAM, "Reason", required=False, profile=COMFORT)
    trade_tax = MultiField(CategoryTradeTax, required=False, profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeAllowanceCharge"



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import MultiStringField, StringField


class IncludedNote(Element):
    content_code = StringField(NS_RAM, "ContentCode", required=False,
                               profile=EXTENDED)
    content = MultiStringField(NS_RAM, "Content", required=False,
                               profile=BASIC)
    subject_code = StringField(NS_RAM, "SubjectCode", required=False,
                               profile=COMFORT)

    class Meta:
        namespace = NS_RAM
        tag = "IncludedNote"



import logging
import os
from xml.dom import minidom

logger = logging.getLogger("drafthorse")


def minify(xml):
    try:
        from lxml import etree
        return b"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + etree.tostring(etree.fromstring(xml))
    except ImportError:
        logger.warning("Could not minify output as LXML is not installed.")
        return xml


def prettify(xml, **kwargs):
    try:
        from lxml import etree
    except ImportError:
        reparsed = minidom.parseString(xml)
        return reparsed.toprettyxml(indent="\t")
    else:
        parser = etree.XMLParser(remove_blank_text=True, **kwargs)
        return b"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + etree.tostring(
            etree.fromstring(xml, parser), pretty_print=True
        )


def validate_xml(xmlout, schema):
    try:
        from lxml import etree
    except ImportError:
        logger.warning("Could not validate output as LXML is not installed.")
        return xmlout
    schema = etree.XMLSchema(file=os.path.join(os.path.dirname(__file__), 'schema', schema + '.xsd'))
    parser = etree.XMLParser(schema=schema)
    xml_root = etree.fromstring(xmlout, parser)
    return b"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + etree.tostring(xml_root, pretty_print=True)



from . import COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import (ClassificationField, IDField, MultiField, QuantityField,
                     StringField)


class ProductCharacteristic(Element):
    type_code = StringField(NS_RAM, "TypeCode", required=True, profile=EXTENDED,
                            _d="Art der Produkteigenschaft")
    description = StringField(NS_RAM, "Description", required=True, profile=EXTENDED)
    value_measure = QuantityField(NS_RAM, "ValueMeasure", required=False,
                                  profile=EXTENDED, _d="Numerische Messgröße")
    value = StringField(NS_RAM, "Value", required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableProductCharacteristic"


class ProductClassification(Element):
    class_code = ClassificationField(NS_RAM, "ClassCode", required=True,
                                     profile=EXTENDED)
    value = StringField(NS_RAM, "ClassName", required=True, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "DesignatedProductClassification"


class OriginCountry(Element):
    id = StringField(NS_RAM, "ID", required=True, profile=EXTENDED,
                     _d="Land der Produktherkunft")

    class Meta:
        namespace = NS_RAM
        tag = "OriginTradeCountry"


class ReferencedProduct(Element):
    global_id = IDField(NS_RAM, "GlobalID", required=False, profile=EXTENDED)
    seller_assigned_id = StringField(NS_RAM, "SellerAssignedID", required=False,
                                     profile=EXTENDED)
    buyer_assigned_id = StringField(NS_RAM, "BuyerAssignedID", required=False,
                                    profile=EXTENDED)
    name = StringField(NS_RAM, "Name", required=False, profile=EXTENDED)
    description = StringField(NS_RAM, "Description", required=False, profile=EXTENDED)
    unit_quantity = QuantityField(NS_RAM, "UnitQuantity", required=False,
                                  profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "IncludedReferencedProduct"


class TradeProduct(Element):
    global_id = IDField(NS_RAM, "GlobalID", required=False, profile=COMFORT)
    seller_assigned_id = StringField(NS_RAM, "SellerAssignedID", required=False,
                                     profile=COMFORT)
    buyer_assigned_id = StringField(NS_RAM, "BuyerAssignedID", required=False,
                                    profile=COMFORT)
    name = StringField(NS_RAM, "Name", required=False, profile=COMFORT)
    description = StringField(NS_RAM, "Description", required=False, profile=COMFORT)
    characteristics = MultiField(ProductCharacteristic, required=False, profile=EXTENDED)
    classifications = MultiField(ProductClassification, required=False, profile=EXTENDED)
    origins = MultiField(OriginCountry, required=False, profile=EXTENDED)
    included_products = MultiField(ReferencedProduct, required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeProduct"



import sys

from codecs import open
from os import path
from setuptools import find_packages, setup

from drafthorse import version

here = path.abspath(path.dirname(__file__))

try:
    # Get the long description from the relevant file
    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
except:  # noqa
    long_description = ''

setup(
    name='drafthorse',
    version=version,
    description='Python ZUGFeRD XML implementation',
    long_description=long_description,
    url='https://github.com/pretix/pretix-drafthorse',
    author='Raphael Michel',
    author_email='michel@rami.io',
    license='Apache License',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='xml banking sepa',
    install_requires=[
        'lxml',
        'PyPDF2'
    ] + (['isoweek'] if sys.version_info < (3, 6) else []),

    packages=find_packages(include=['drafthorse', 'drafthorse.*']),
    include_package_data=True,
)



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import (AgencyIDField, CurrencyField, DateTimeField, DecimalField,
                     Field, MultiCurrencyField, MultiStringField,
                     QuantityField, StringField)


class PayerFinancialAccount(Element):
    iban = StringField(NS_RAM, "IBANID")
    proprietary_id = StringField(NS_RAM, "ProprietaryID")

    class Meta:
        namespace = NS_RAM
        tag = "PayerPartyDebtorFinancialAccount"


class PayerFinancialInstitution(Element):
    bic = StringField(NS_RAM, "BICID")
    german_blz = StringField(NS_RAM, "GermanBankleitzahlID")
    name = StringField(NS_RAM, "Name")

    class Meta:
        namespace = NS_RAM
        tag = "PayerSpecifiedDebtorFinancialInstitution"


class PayeeFinancialAccount(Element):
    iban = StringField(NS_RAM, "IBANID")
    account_name = StringField(NS_RAM, "AccountName")
    proprietary_id = StringField(NS_RAM, "ProprietaryID")

    class Meta:
        namespace = NS_RAM
        tag = "PayeePartyCreditorFinancialAccount"


class PayeeFinancialInstitution(Element):
    bic = StringField(NS_RAM, "BICID")
    german_blz = StringField(NS_RAM, "GermanBankleitzahlID")
    name = StringField(NS_RAM, "Name")

    class Meta:
        namespace = NS_RAM
        tag = "PayeeSpecifiedCreditorFinancialInstitution"


class PaymentMeans(Element):
    type_code = StringField(NS_RAM, "TypeCode", required=False, profile=COMFORT)
    information = MultiStringField(NS_RAM, "Information", required=False, profile=COMFORT)
    id = AgencyIDField(NS_RAM, "ID", required=False, profile=BASIC)
    payer_account = Field(PayerFinancialAccount)
    payer_institution = Field(PayerFinancialInstitution)
    payee_account = Field(PayeeFinancialAccount)
    payee_institution = Field(PayeeFinancialInstitution)

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradeSettlementPaymentMeans"


class PaymentPenaltyTerms(Element):
    basis_date_time = DateTimeField(NS_RAM, "BasisDateTime", required=False,
                                    profile=EXTENDED, _d="Bezugsdatum der Fälligkeit")
    basis_period_measure = QuantityField(NS_RAM, "BasisPeriodMeasure", required=False,
                                         profile=EXTENDED, _d="Fälligkeitszeitraum")
    basis_amount = CurrencyField(NS_RAM, "BasisAmount", required=False,
                                 profile=EXTENDED, _d="Basisbetrag des Zahlungszuschlags")
    calculation_percent = DecimalField(NS_RAM, "CalculationPercent", required=False,
                                       profile=EXTENDED, _d="Prozentwert des Zahlungszuschlags")
    actual_amount = CurrencyField(NS_RAM, "ActualPenaltyAmount", required=False,
                                  profile=EXTENDED, _d="Betrag des Zahlungszuschlags")

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradePaymentPenaltyTerms"


class PaymentDiscountTerms(Element):
    basis_date_time = DateTimeField(NS_RAM, "BasisDateTime", required=False,
                                    profile=EXTENDED, _d="Bezugsdatum der Fälligkeit")
    basis_period_measure = QuantityField(NS_RAM, "BasisPeriodMeasure", required=False,
                                         profile=EXTENDED, _d="Fälligkeitszeitraum")
    basis_amount = CurrencyField(NS_RAM, "BasisAmount", required=False,
                                 profile=EXTENDED, _d="Basisbetrag des Zahlungsabschlags")
    calculation_percent = DecimalField(NS_RAM, "CalculationPercent", required=False,
                                       profile=EXTENDED, _d="Prozentwert des Zahlungsabschlags")
    actual_amount = CurrencyField(NS_RAM, "ActualDiscountAmount", required=False,
                                  profile=EXTENDED, _d="Betrag des Zahlungsabschlags")

    class Meta:
        namespace = NS_RAM
        tag = "ApplicableTradePaymentDiscountTerms"


class PaymentTerms(Element):
    description = StringField(NS_RAM, "Description", required=True, profile=COMFORT,
                              _d="Freitext der Zahlungsbedingungen")
    due = DateTimeField(NS_RAM, "DueDateDateTime", required=False, profile=COMFORT,
                        _d="Fälligkeitsdatum")
    partial_amount = MultiCurrencyField(NS_RAM, "PartialPaymentAmount", profile=EXTENDED,
                                        required=False, _d="Betrag der Teilzahlung")
    penalty_terms = Field(PaymentPenaltyTerms, required=False, profile=EXTENDED,
                          _d="Detailinformationen zu Zahlungszuschlägen")
    discount_terms = Field(PaymentDiscountTerms, required=False, profile=EXTENDED,
                           _d="Detailinformationen zu Zahlungsabschlägen")

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTradePaymentTerms"



from . import BASIC, COMFORT, EXTENDED, NS_RAM
from .elements import Element
from .fields import Field, IDField, MultiField, MultiIDField, StringField


class PostalTradeAddress(Element):
    postcode = StringField(NS_RAM, "PostcodeCode", required=False, profile=BASIC)
    line_one = StringField(NS_RAM, "LineOne", required=False, profile=BASIC)
    line_two = StringField(NS_RAM, "LineTwo", required=False, profile=BASIC)
    city_name = StringField(NS_RAM, "CityName", required=False, profile=BASIC)
    country_id = StringField(NS_RAM, "CountryID", required=False, profile=BASIC)

    class Meta:
        namespace = NS_RAM
        tag = "PostalTradeAddress"


class TaxRegistration(Element):
    id = IDField(NS_RAM, "ID")

    class Meta:
        namespace = NS_RAM
        tag = "SpecifiedTaxRegistration"


class PhoneNumber(Element):
    number = StringField(NS_RAM, "CompleteNumber", required=False,
                         profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "TelephoneUniversalCommunication"


class FaxNumber(Element):
    number = StringField(NS_RAM, "CompleteNumber", required=False,
                         profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "FaxUniversalCommunication"


class EmailURI(Element):
    address = StringField(NS_RAM, "URIID", required=False,
                          profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "EmailURIUniversalCommunication"


class TradeContact(Element):
    person_name = StringField(NS_RAM, "PersonName", required=False,
                              profile=EXTENDED)
    department_name = StringField(NS_RAM, "DepartmentName", required=False,
                                  profile=EXTENDED)
    telephone = Field(PhoneNumber, required=False, profile=EXTENDED)
    fax = Field(FaxNumber, required=False, profile=EXTENDED)
    email = Field(EmailURI, required=False, profile=EXTENDED)

    class Meta:
        namespace = NS_RAM
        tag = "DefinedTradeContact"


class TradeParty(Element):
    id = StringField(NS_RAM, "ID", required=False, profile=COMFORT,
                     _d="Identifier des Verkäufers")
    global_id = MultiIDField(NS_RAM, "GlobalID", required=False, profile=COMFORT,
                             _d="Globaler Identifier des Verkäufers")
    name = StringField(NS_RAM, "Name", required=False, profile=BASIC)
    contact = Field(TradeContact, required=False, profile=EXTENDED,
                    _d="Ansprechpartner des Käufers")
    address = Field(PostalTradeAddress, required=False, profile=BASIC,
                    _d="Anschrift des Käufers")
    tax_registrations = MultiField(TaxRegistration, required=False, profile=BASIC)


class PayeeTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "PayeeTradeParty"


class InvoiceeTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "InvoiceeTradeParty"


class BuyerTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "BuyerTradeParty"


class SellerTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "SellerTradeParty"


class EndUserTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "ProductEndUserTradeParty"


class ShipToTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "ShipToTradeParty"


class ShipFromTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "ShipFromTradeParty"


class UltimateShipToTradeParty(TradeParty):
    class Meta:
        namespace = NS_RAM
        tag = "UltimateShipToTradeParty"



from . import BASIC
from .container import (Container, CurrencyContainer, IDContainer,
                        StringContainer)


class Field:
    def __init__(self, cls, default=False, required=False, profile=BASIC, _d=None):
        self.cls = cls
        self.required = required
        self.default = default
        self.__doc__ = _d
        self.profile = profile
        super().__init__()

    def initialize(self):
        return self.cls(required=self.required)

    def __get__(self, instance, objtype):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        return instance._data[self.name]

    def __set__(self, instance, value):
        raise AttributeError("Read-only!")

    def __delete__(self, instance):
        del instance._data[self.name]

    def __set_name__(self, owner, name):
        self.name = name


class StringField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import StringElement
        super().__init__(StringElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name].text = value


class AgencyIDField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import AgencyIDElement
        super().__init__(AgencyIDElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 2-tuple of scheme agency ID and ID.")
        instance._data[self.name].text = value[1]
        instance._data[self.name].scheme_id = value[0]


class ClassificationField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import ClassificationElement
        super().__init__(ClassificationElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 3-tuple of list ID, list version and ID.")
        instance._data[self.name].text = value[2]
        instance._data[self.name].list_id = value[0]
        instance._data[self.name].list_version_id = value[1]


class IDField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import IDElement
        super().__init__(IDElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(self.namespace, self.tag)

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 2-tuple of including scheme ID and ID.")
        instance._data[self.name].text = value[1]
        instance._data[self.name].scheme_id = value[0]


class DecimalField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import DecimalElement
        super().__init__(DecimalElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name].value = value

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class QuantityField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import QuantityElement
        super().__init__(QuantityElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 2-tuple of including amount and unit code.")
        instance._data[self.name].amount = value[0]
        instance._data[self.name].unit_code = value[1]

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class CurrencyField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import CurrencyElement
        super().__init__(CurrencyElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()

        if not isinstance(value, (tuple, list)):
            raise TypeError("Please pass a 2-tuple of including amount and currency.")
        instance._data[self.name].amount = value[0]
        instance._data[self.name].currency = value[1]

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class IndicatorField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import IndicatorElement
        super().__init__(IndicatorElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name].value = value

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class DateTimeField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import DateTimeElement
        super().__init__(DateTimeElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name].value = value

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class DirectDateTimeField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        from .elements import DirectDateTimeElement
        super().__init__(DirectDateTimeElement, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def __set__(self, instance, value):
        if instance._data.get(self.name, None) is None:
            instance._data[self.name] = self.initialize()
        instance._data[self.name].value = value

    def initialize(self):
        return self.cls(self.namespace, self.tag)


class MultiField(Field):
    def __init__(self, inner_type, default=False, required=False, profile=BASIC, _d=None):
        super().__init__(Container, default, required, profile, _d)
        self.inner_type = inner_type

    def initialize(self):
        return self.cls(child_type=self.inner_type)


class MultiStringField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        super().__init__(StringContainer, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(child_type=str, namespace=self.namespace, tag=self.tag)


class MultiCurrencyField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        super().__init__(CurrencyContainer, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(child_type=(tuple, list), namespace=self.namespace, tag=self.tag)


class MultiIDField(Field):
    def __init__(self, namespace, tag, default=False, required=False, profile=BASIC, _d=None):
        super().__init__(IDContainer, default, required, profile, _d)
        self.namespace = namespace
        self.tag = tag

    def initialize(self):
        return self.cls(child_type=(tuple, list), namespace=self.namespace, tag=self.tag)
